# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
import scipy.stats as st

from dscontrib.flawrence.abtest_stats import (
    res_columns, one_res_index, compare_two_sample_sets
)


def compare_two(
    df, col_label, focus_label=None, control_label='control', num_samples=10000
):
    """Jointly sample conversion rates for two branches then compare them.

    See `compare_two_from_summary` for more details.

    Args:
        df: a pandas DataFrame of experiment data. Each row represents
            data about an individual test subject. One column is named
            'branch' and contains the test subject's branch. The other
            columns contain the test subject's values for each metric.
            The column to be analyzed (named `col_label`) should be
            boolean or 0s and 1s.
        col_label: Label for the df column contaning the metric to be
            analyzed.
        focus_label: String in `df['branch']` that identifies the
            target non-control branch, the branch for which we want to
            calculate uplifts.
        control_label: String in `df['branch']` that identifies the
            control branch, the branch with respect to which we want to
            calculate uplifts.
        num_samples: The number of samples to compute

    Returns a pandas.Series of summary statistics for the possible
    uplifts - see docs for `compare_two_sample_sets`
    """
    # I would have used `isin` but it seems to be ~100x slower?
    assert ((df[col_label] == 0) | (df[col_label] == 1)).all()

    summary = df.groupby('branch')[col_label].agg({
        'num_enrollments': len,
        'num_conversions': np.sum
    })

    if len(summary) > 2:
        # Multi-branch test; need to know which two branches to compare
        assert focus_label is not None
        summary = summary.loc[[focus_label, control_label]]

    return compare_two_from_summary(
        summary, control_label=control_label, num_samples=num_samples
    )


def compare_many(df, col_label, num_samples=10000):
    """Jointly sample conversion rates for many branches then compare them.

    See `compare_many_from_summary` for more details.

    Args:
        df: a pandas DataFrame of experiment data. Each row represents
            data about an individual test subject. One column is named
            'branch' and contains the test subject's branch. The other
            columns contain the test subject's values for each metric.
            The column to be analyzed (named `col_label`) should be
            boolean or 0s and 1s.
        col_label: Label for the df column contaning the metric to be
            analyzed.
        num_samples: The number of samples to compute

    Returns a pandas.DataFrame of summary statistics for the possible
    uplifts:
        - columns: equivalent to rows output by `compare_two()`
        - index: list of branches
    """
    assert (df[col_label] == 0) | (df[col_label] == 1).all()
    summary = df.groupby('branch')[col_label].agg({
        'num_enrollments': len,
        'num_conversions': np.sum
    })

    return compare_many_from_summary(
        summary, num_samples=num_samples
    )


def summarize_one_from_summary(
    s, num_enrollments_label='num_enrollments', num_conversions_label='num_conversions'
):
    res = pd.Series(index=one_res_index)
    res['mean'] = s.loc[num_conversions_label] / s.loc[num_enrollments_label]

    ppfs = [0.005, 0.05, 0.95, 0.995]
    res[[str(v) for v in ppfs]] = st.beta(
        s.loc[num_conversions_label] + 1,
        s.loc[num_enrollments_label] - s.loc[num_conversions_label] + 1
    ).ppf(ppfs)

    return res


def compare_two_from_summary(
    df,
    control_label='control',
    num_enrollments_label='num_enrollments',
    num_conversions_label='num_conversions',
    num_samples=10000
):
    """Jointly sample conversion rates for two branches then compare them.

    Calculates various quantiles on the uplift of the non-control
    branch's sampled conversion rates with respect to the control
    branch's sampled conversion rates.

    The data in `df` is modelled as being generated binomially, with a
    Beta(1, 1) (uniform) prior over the conversion rate parameter.

    Args:
        df: A pandas dataframe of integers:
            - df.index lists the two experiment branches
            - df.columns is
                (num_enrollments_label, num_conversions_label)
        control_label: Label for the df row containing data for the
            control branch
        num_enrollments_label: Label for the df column containing the
            number of enrollments in each branch.
        num_conversions_label: Label for the df column containing the
            number of conversions in each branch.
        num_samples: The number of samples to compute

    Returns a pandas.Series of summary statistics for the possible
    uplifts - see docs for `compare_two_sample_sets`
    FIXME: update docs
    """
    assert len(df.index) == 2
    assert control_label in df.index, "Which branch is the control?"

    test_label = list(set(df.index) - {control_label})[0]

    samples = _generate_samples(
        df, num_enrollments_label, num_conversions_label, num_samples
    )

    comparative = compare_two_sample_sets(samples[test_label], samples[control_label])
    comparative.name = num_conversions_label

    individual = {
        l: summarize_one_from_summary(
            df.loc[l], num_enrollments_label, num_conversions_label
        ) for l in [control_label, test_label]
    }

    return {
        'comparative': comparative,
        'individual': individual
    }


def compare_many_from_summary(
    df,
    num_enrollments_label='num_enrollments',
    num_conversions_label='num_conversions',
    num_samples=10000
):
    """Jointly sample conversion rates for many branches then compare them.

    Calculates various quantiles on the uplift of each branch's sampled
    conversion rates, with respect to the best of the other branches'
    sampled conversion rates.

    The data in `df` is modelled as being generated binomially, with a
    Beta(1, 1) (uniform) prior over the conversion rate parameter.

    Args:
        df: A pandas dataframe of integers:
            - df.index lists the experiment branches
            - df.columns is
                (num_enrollments_label, num_conversions_label)
        control_label: Label for the df row containing data for the
            control branch
        num_enrollments_label: Label for the df column containing the
            number of enrollments in each branch.
        num_conversions_label: Label for the df column containing the
            number of conversions in each branch.
        num_samples: The number of samples to compute

    Returns a pandas.DataFrame of summary statistics for the possible
    uplifts:
        - columns: equivalent to rows output by `compare_two()`
        - index: list of branches
    """
    samples = _generate_samples(
        df, num_enrollments_label, num_conversions_label, num_samples
    )

    comparative = pd.DataFrame(index=df.index, columns=res_columns)
    comparative.name = num_conversions_label

    individual = {}

    for branch in df.index:
        # Compare this branch to the best of the rest
        # (beware Monty's Revenge!)
        this_branch = samples[branch]
        # Warning: assumes we're trying to maximise the metric
        best_of_rest = samples.drop(branch, axis='columns').max(axis='columns')

        comparative.loc[branch] = compare_two_sample_sets(this_branch, best_of_rest)
        individual[branch] = summarize_one_from_summary(
            df.loc[branch], num_enrollments_label, num_conversions_label
        )

    return {
        'comparative': comparative,
        'individual': individual
    }


def _generate_samples(
    df, num_enrollments_label, num_conversions_label, num_samples
):
    """Return samples from Beta distributions.

    Assumes a Beta(1, 1) prior.

        Args:
        df: A pandas dataframe of integers:
            - df.index lists the experiment branches
            - df.columns is
                (num_enrollments_label, num_conversions_label)
        num_enrollments_label: Label for the df column containing the
            number of enrollments in each branch.
        num_conversions_label: Label for the df column containing the
            number of conversions in each branch.
        num_samples: The number of samples to compute

    Returns a pandas.DataFrame of sampled conversion rates:
        - columns: list of branches
        - index: enumeration of samples
    """
    samples = pd.DataFrame(index=np.arange(num_samples), columns=df.index)
    for branch_label, r in df.iterrows():
        # Oh, for a prior...
        samples[branch_label] = np.random.beta(
            r.loc[num_conversions_label] + 1,
            r.loc[num_enrollments_label] - r.loc[num_conversions_label] + 1,
            size=num_samples
        )

    return samples



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd


# TODO: is this kind of standardization helpful?
# Will it discourage Bonferroni corrections? Is that a major problem?

res_columns = [
    'rel_uplift_0.005',
    'rel_uplift_0.05',
    'rel_uplift_0.5',
    'rel_uplift_0.95',
    'rel_uplift_0.995',
    'rel_uplift_exp',
    'abs_uplift_0.005',
    'abs_uplift_0.05',
    'abs_uplift_0.5',
    'abs_uplift_0.95',
    'abs_uplift_0.995',
    'abs_uplift_exp',
    'max_abs_diff_0.95',
]


one_res_index = [
    '0.005',
    '0.05',
    'mean',
    '0.95',
    '0.995',
]

# TODO: use dictionaries instead of Series?


def compare_two_sample_sets(focus, reference):
    res = pd.Series(index=res_columns)

    rel_uplift_samples = focus / reference - 1

    res[
        'rel_uplift_0.005':'rel_uplift_0.995'
    ] = np.quantile(rel_uplift_samples, [0.005, 0.05, 0.5, 0.95, 0.995])
    res['rel_uplift_exp'] = np.mean(rel_uplift_samples)

    abs_uplift_samples = focus - reference

    res[
        'abs_uplift_0.005':'abs_uplift_0.995'
    ] = np.quantile(abs_uplift_samples, [0.005, 0.05, 0.5, 0.95, 0.995])
    res['abs_uplift_exp'] = np.mean(abs_uplift_samples)

    res['max_abs_diff_0.95'] = np.quantile(np.abs(abs_uplift_samples), 0.95)

    res['prob_win'] = np.mean(focus > reference)

    return res


def summarize_one_sample_set(data):
    res = pd.Series(index=['0.05', 'mean', '0.95'])
    res[['0.05', '0.95']] = np.quantile(data, [0.05, 0.95])
    res['mean'] = np.mean(data)
    return res



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import datetime


def add_days(date_string, n_days):
    """Add `n_days` days to a date string like '20190101'."""
    # TODO: would love to do some pre-computation here!
    original_date = datetime.datetime.strptime(date_string, '%Y%m%d')
    new_date = original_date + datetime.timedelta(days=n_days)
    return datetime.datetime.strftime(new_date, '%Y%m%d')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np

import mozanalysis.stats as mzas

from dscontrib.flawrence.abtest_stats import (
    summarize_one_sample_set, compare_two_sample_sets
)


def bootstrap_one(sc, data, num_samples=10000, seed_start=0):
    """Bootstrap on the means of one variation on its own.

    Generates `num_samples` sampled means, then returns summary
    statistics for their distribution.

    Args:
        sc: The spark context
        data: The data as a list, 1D numpy array, or pandas Series
        num_samples: The number of bootstrap iterations to perform
        seed_start: An int with which to seed numpy's RNG. It must
            be unique to this set of calculations.
    """
    samples = _resample_parallel(sc, data, num_samples, seed_start)
    return summarize_one_sample_set(samples)


def bootstrap_two(
    sc,
    df=None, col_label=None, focus_label=None, control_label='control',
    focus=None, reference=None,
    num_samples=10000, filter_outliers=None
):
    """Jointly sample bootstrapped means from two distributions then compare them.

    Calculates various quantiles on the uplift of the focus branch's
    mean value with respect to the reference branch's mean value.

    Either supply df, col_label focus_label and optionally control_label,
    or supply focus and reference.

    Args:
        sc: The spark context
        df: a pandas DataFrame of experiment data. Each row represents
            data about an individual test subject. One column is named
            'branch' and contains the test subject's branch. The other
            columns contain the test subject's values for each metric.
        col_label: When df is supplied, the name of the column
            containing the metric of interest.
        control_label: When df is supplied, the name of the control
            branch.
        focus: When df is not supplied, the data for the focal branch
            as a list, 1D numpy array, or pandas Series.
        reference: When df is not supplied, the data for the reference
            (typically control) branch.
        num_samples: The number of bootstrap iterations to perform
        filter_outliers: An optional threshold quantile, above which to
            discard outliers.

    Returns a dictionary:
        'comparative': pandas.Series of summary statistics for the possible
            uplifts - see docs for `compare_two_sample_sets`
        'individual': list of summary stats for (focus, reference) means.
            Each set of summary stats is a pandas.Series
    """
    if df is None:
        assert focus is not None and reference is not None
        assert col_label is None and focus_label is None
    else:
        assert focus is None and reference is None
        assert focus_label is not None
        focus = df[col_label][df.branch == focus_label]
        reference = df[col_label][df.branch == control_label]

    # FIXME: should we be filtering or truncating outliers?
    if filter_outliers:
        assert filter_outliers < 1
        focus = focus[focus <= np.quantile(focus, filter_outliers)]
        reference = reference[reference <= np.quantile(reference, filter_outliers)]

    focus_samples = _resample_parallel(sc, focus, num_samples)
    reference_samples = _resample_parallel(sc, reference, num_samples)

    return {
        'comparative':
            compare_two_sample_sets(focus_samples, reference_samples),
        'individual': [
                summarize_one_sample_set(focus_samples),
                summarize_one_sample_set(reference_samples)
            ],
    }


def _resample_parallel(sc, data, num_samples, seed_start=None):
    """Return bootstrapped samples for the mean of `data`.

    Do the resampling in parallel over the cluster.

    Args:
        sc: The spark context
        data: The data as a list, numpy array, or pandas series
        num_samples: The number of samples to return
        seed_start: A seed for the random number generator; this
            function will use seeds in the range
                [seed_start, seed_start + num_samples)
            and these particular seeds must not be used elsewhere
            in this calculation. By default, use a random seed.

    Returns a numpy array of sampled means
    """
    if not type(data) == np.ndarray:
        data = np.array(data)

    if seed_start is None:
        seed_start = np.random.randint(np.iinfo(np.uint32).max)

    # Deterministic "randomness" requires careful state handling :(
    # Need to ensure every iteration has a unique, deterministic seed.
    # N.B. `num_iterations` in mozanalysis is called `num_samples` here
    seed_range = range(seed_start, seed_start + num_samples)

    try:
        broadcast_data = sc.broadcast(data)

        summary_stat_samples = sc.parallelize(seed_range).map(
            lambda seed: mzas._resample(
                iteration=seed % np.iinfo(np.uint32).max,
                stat_fn=np.mean,
                broadcast_data=broadcast_data,
            )
        ).collect()

        return np.array(summary_stat_samples)

    finally:
        broadcast_data.unpersist()


def _resample_local(data, num_samples):
    """Equivalent to `_resample_parallel` but doesn't require Spark.

    The main purpose of this function is to document what's being done
    in `_resample_parallel` :D
    """
    return np.array([
        np.mean(np.random.choice(data, size=len(data)))
        for _ in range(num_samples)
    ])
